# -*- coding: utf-8 -*-
from plone.namedfile.file import NamedBlobFile as BlobFileValueType
from plone.namedfile.file import NamedBlobImage as BlobImageValueType
from plone.namedfile.file import NamedFile as FileValueType
from plone.namedfile.file import NamedImage as ImageValueType
from plone.namedfile.interfaces import INamedBlobFile
from plone.namedfile.interfaces import INamedBlobFileField
from plone.namedfile.interfaces import INamedBlobImage
from plone.namedfile.interfaces import INamedBlobImageField
from plone.namedfile.interfaces import INamedFile
from plone.namedfile.interfaces import INamedFileField
from plone.namedfile.interfaces import INamedImage
from plone.namedfile.interfaces import INamedImageField
from plone.namedfile.utils import get_contenttype
from zope.i18nmessageid import MessageFactory
from zope.interface import implementer
from zope.schema import Object
from zope.schema import ValidationError


_ = MessageFactory('plone')


class InvalidImageFile(ValidationError):
    """Exception for invalid image file"""
    __doc__ = _(u'Invalid image file')


def validate_image_field(field, value):
    if value is not None:
        mimetype = get_contenttype(value)
        if mimetype.split('/')[0] != 'image':
            raise InvalidImageFile(mimetype, field.__name__)


@implementer(INamedFileField)
class NamedFile(Object):
    """A NamedFile field
    """

    _type = FileValueType
    schema = INamedFile

    def __init__(self, **kw):
        if 'schema' in kw:
            self.schema = kw.pop('schema')
        super(NamedFile, self).__init__(schema=self.schema, **kw)


@implementer(INamedImageField)
class NamedImage(Object):
    """A NamedImage field
    """

    _type = ImageValueType
    schema = INamedImage

    def __init__(self, **kw):
        if 'schema' in kw:
            self.schema = kw.pop('schema')
        super(NamedImage, self).__init__(schema=self.schema, **kw)

    def _validate(self, value):
        super(NamedImage, self)._validate(value)
        validate_image_field(self, value)


@implementer(INamedBlobFileField)
class NamedBlobFile(Object):
    """A NamedBlobFile field
    """

    _type = BlobFileValueType
    schema = INamedBlobFile

    def __init__(self, **kw):
        if 'schema' in kw:
            self.schema = kw.pop('schema')
        super(NamedBlobFile, self).__init__(schema=self.schema, **kw)


@implementer(INamedBlobImageField)
class NamedBlobImage(Object):
    """A NamedBlobImage field
    """

    _type = BlobImageValueType
    schema = INamedBlobImage

    def __init__(self, **kw):
        if 'schema' in kw:
            self.schema = kw.pop('schema')
        super(NamedBlobImage, self).__init__(schema=self.schema, **kw)

    def _validate(self, value):
        super(NamedBlobImage, self)._validate(value)
        validate_image_field(self, value)



# -*- coding: utf-8 -*-
# This file was borrowed from z3c.blobfile and is licensed under the terms of
# the ZPL.
from OFS.Image import Pdata
from plone.namedfile.file import FileChunk
from plone.namedfile.interfaces import IStorage
from plone.namedfile.interfaces import NotStorable
from zope.interface import implementer
from zope.publisher.browser import FileUpload


MAXCHUNKSIZE = 1 << 16


@implementer(IStorage)
class StringStorable(object):

    def store(self, data, blob):
        if not isinstance(data, str):
            raise NotStorable('Could not store data (not of "str" type).')

        with blob.open('w') as fp:
            fp.write(data)


@implementer(IStorage)
class UnicodeStorable(StringStorable):

    def store(self, data, blob):
        if not isinstance(data, unicode):
            raise NotStorable('Could not store data (not of "unicode" type).')

        data = data.encode('UTF-8')
        StringStorable.store(self, data, blob)


@implementer(IStorage)
class FileChunkStorable(object):

    def store(self, data, blob):
        if not isinstance(data, FileChunk):
            raise NotStorable('Could not store data (not a of "FileChunk" type).')  # noqa

        with blob.open('w') as fp:
            chunk = data
            while chunk:
                fp.write(chunk._data)
                chunk = chunk.next


@implementer(IStorage)
class FileDescriptorStorable(object):

    def store(self, data, blob):
        if not isinstance(data, file):
            raise NotStorable('Could not store data (not of "file").')

        filename = getattr(data, 'name', None)
        if filename is not None:
            blob.consumeFile(filename)
            return


@implementer(IStorage)
class FileUploadStorable(object):

    def store(self, data, blob):
        if not isinstance(data, FileUpload):
            raise NotStorable('Could not store data (not of "FileUpload").')

        data.seek(0)

        with blob.open('w') as fp:
            block = data.read(MAXCHUNKSIZE)
            while block:
                fp.write(block)
                block = data.read(MAXCHUNKSIZE)


@implementer(IStorage)
class PDataStorable(object):
    def store(self, pdata, blob):
        if not isinstance(pdata, Pdata):
            raise NotStorable('Could not store data (not of "Pdata").')
        fp = blob.open('w')
        fp.write(str(pdata))
        fp.close()



# -*- coding: utf-8 -*-

from logging import getLogger
from plone.namedfile.interfaces import IBlobby
from plone.namedfile.utils.jpeg_utils import process_jpeg
from plone.namedfile.utils.png_utils import process_png
from plone.namedfile.utils.tiff_utils import process_tiff
from plone.registry.interfaces import IRegistry
from StringIO import StringIO
from zope.component import queryUtility
from zope.deprecation import deprecate

import mimetypes
import os.path
import piexif
import PIL.Image
import struct
import urllib


log = getLogger(__name__)

try:
    from Products.CMFPlone.interfaces.controlpanel import IImagingSchema
except ImportError:
    IImagingSchema = None
    log.info('IImagingSchema for high pixel density scales not available.')


try:
    # use this to stream data if we can
    from ZPublisher.Iterators import filestream_iterator
except ImportError:
    filestream_iterator = None


def safe_basename(filename):
    """Get the basename of the given filename, regardless of which platform
    (Windows or Unix) it originated from.
    """
    fslice = max(
        filename.rfind('/'),
        filename.rfind('\\'),
        filename.rfind(':'),
    ) + 1
    return filename[fslice:]


def get_contenttype(
        file=None,
        filename=None,
        default='application/octet-stream'):
    """Get the MIME content type of the given file and/or filename.
    """

    file_type = getattr(file, 'contentType', None)
    if file_type:
        return file_type

    filename = getattr(file, 'filename', filename)
    if filename:
        extension = os.path.splitext(filename)[1].lower()
        return mimetypes.types_map.get(extension, 'application/octet-stream')

    return default


def set_headers(file, response, filename=None):
    """Set response headers for the given file. If filename is given, set
    the Content-Disposition to attachment.
    """

    contenttype = get_contenttype(file)

    response.setHeader('Content-Type', contenttype)
    response.setHeader('Content-Length', file.getSize())

    if filename is not None:
        if not isinstance(filename, unicode):
            filename = unicode(filename, 'utf-8', errors='ignore')
        filename = urllib.quote(filename.encode('utf8'))
        response.setHeader(
            'Content-Disposition',
            'attachment; filename*=UTF-8\'\'{0}'.format(filename)
        )


def stream_data(file):
    """Return the given file as a stream if possible.
    """

    if IBlobby.providedBy(file):
        if file._blob._p_blob_uncommitted:
            return file.data
        if filestream_iterator is not None:
            return filestream_iterator(file._blob.committed(), 'rb')

    return file.data


def _ensure_data(image):
    data = None
    if getattr(image, 'read', None):
        data = image.read()
        image.seek(0)
    else:
        data = image
    return str(data)


def getImageInfo(data):
    data = _ensure_data(data)
    size = len(data)
    height = -1
    width = -1
    content_type = ''
    if (size >= 10) and data[:6] in ('GIF87a', 'GIF89a'):
        # handle GIFs
        content_type = 'image/gif'
        w, h = struct.unpack('<HH', data[6:10])
        width = int(w)
        height = int(h)

    elif data[:8] == '\211PNG\r\n\032\n':
        # handle PNG
        content_type, width, height = process_png(data)

    elif data[:2] == '\377\330':
        # handle JPEGs
        content_type, width, height = process_jpeg(data)

    elif (size >= 30) and data.startswith('BM'):
        # handle BMPs
        kind = struct.unpack('<H', data[14:16])[0]
        if kind == 40:  # Windows 3.x bitmap
            content_type = 'image/x-ms-bmp'
            width, height = struct.unpack('<LL', data[18:26])

    elif (size >= 4) and data[:4] in ['MM\x00*', 'II*\x00']:
        # handle TIFFs
        content_type, width, height = process_tiff(data)
    elif size >= 4 and data[:5] == '<?xml' and '<svg' in data:
        # handle SVG
        content_type = 'image/svg+xml'
        log.info('Can not handle SVG. Image sizes are not set properly.')
    else:
        # Use PIL / Pillow to determ Image Information
        try:
            img = PIL.Image.open(StringIO(data))
            width, height = img.size
            content_type = img.format
        except Exception:
            # TODO: determ wich error really happens
            # Should happen if data is to short --> first_bytes
            # happens also if data is an svg or another special format.
            log.warn(
                'PIL can not recognize the image. '
                'Image is probably broken or of a non-supported format.'
            )

    log.debug('Image Info (Type: %s, Width: %s, Height: %s)',
              content_type, width, height)
    return content_type, width, height


def get_exif(image):
    #
    exif_data = None
    image_data = _ensure_data(image)

    content_type, width, height = getImageInfo(image_data)
    if content_type in ['image/jpeg', 'image/tiff']:
        # Only this two Image Types could have Exif informations
        # see http://www.cipa.jp/std/documents/e/DC-008-2012_E.pdf
        try:
            exif_data = piexif.load(image_data)
        except Exception as e:
            # TODO: determ wich error really happens
            # Should happen if data is to short --> first_bytes
            log.warn(e)
            exif_data = exif_data = {
                '0th': {
                    piexif.ImageIFD.XResolution: (width, 1),
                    piexif.ImageIFD.YResolution: (height, 1),
                }
            }
    return exif_data


def rotate_image(image_data, method=None, REQUEST=None):
    """Rotate Image if it has Exif Orientation Informations other than 1.

    Do not use PIL.Image.rotate function as this did not transpose the image,
    rotate keeps the image width and height and rotates the image around a
    central point. PIL.Image.transpose also changes Image Orientation.
    """
    orientation = 1  # if not set assume correct orrinetation --> 1
    data = _ensure_data(image_data)
    img = PIL.Image.open(StringIO(data))

    exif_data = None
    if 'exif' in img.info:
        try:
            exif_data = piexif.load(img.info['exif'])
        except ValueError:
            log.warn('Exif information currupt')
            pass
        if exif_data and piexif.ImageIFD.Orientation in exif_data['0th']:
            orientation = exif_data['0th'][piexif.ImageIFD.Orientation]
        if exif_data and \
                (not exif_data['0th'].get(piexif.ImageIFD.XResolution) or
                 not exif_data['0th'].get(piexif.ImageIFD.YResolution)):
            exif_data['0th'][piexif.ImageIFD.XResolution] = (img.width, 1)
            exif_data['0th'][piexif.ImageIFD.YResolution] = (img.height, 1)
    if exif_data is None:
        width, height = img.size
        exif_data = {
            '0th': {
                piexif.ImageIFD.XResolution: (width, 1),
                piexif.ImageIFD.YResolution: (height, 1),
            }
        }

    if method is not None:
        orientation = method

    log.debug('Rotate image with input orientation: %s', orientation)

    fmt = img.format
    if orientation == 1:  # not transform necessary
        # img = img
        pass
    elif orientation == 2:
        img = img.transpose(PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 3:
        img = img.transpose(PIL.Image.ROTATE_180)
    elif orientation == 4:
        img = img.transpose(PIL.Image.ROTATE_180).transpose(
            PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 5:
        img = img.transpose(PIL.Image.ROTATE_270).transpose(
            PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 6:
        img = img.transpose(PIL.Image.ROTATE_270)
    elif orientation == 7:
        img = img.transpose(PIL.Image.ROTATE_90).transpose(
            PIL.Image.FLIP_LEFT_RIGHT)
    elif orientation == 8:
        img = img.transpose(PIL.Image.ROTATE_90)

    if orientation in [5, 6, 7, 8]:
        if exif_data['0th'][piexif.ImageIFD.XResolution] and \
                exif_data['0th'][piexif.ImageIFD.YResolution]:
            exif_data['0th'][piexif.ImageIFD.XResolution], \
                exif_data['0th'][piexif.ImageIFD.YResolution] = \
                exif_data['0th'][piexif.ImageIFD.YResolution], \
                exif_data['0th'][piexif.ImageIFD.XResolution]
        else:
            exif_data['0th'][piexif.ImageIFD.XResolution], \
                exif_data['0th'][piexif.ImageIFD.YResolution] = \
                (img.width, 1), (img.height, 1)

    # set orientation to normal
    exif_data['0th'][piexif.ImageIFD.Orientation] = 1

    try:
        exif_bytes = piexif.dump(exif_data)
    except Exception as e:
        log.warn(e)
        del(exif_data['Exif'][piexif.ExifIFD.SceneType])
        # This Element piexif.ExifIFD.SceneType cause error on dump
        exif_bytes = piexif.dump(exif_data)
    output_image_data = StringIO()
    img.save(output_image_data, format=fmt, exif=exif_bytes)
    width, height = img.size
    return output_image_data.getvalue(), width, height, exif_data


@deprecate('use getHighPixelDensityScales instead')
def getRetinaScales():
    return getHighPixelDensityScales()


def getHighPixelDensityScales():
    registry = queryUtility(IRegistry)
    if IImagingSchema and registry:
        settings = registry.forInterface(
            IImagingSchema, prefix='plone', check=False)
        if settings.highpixeldensity_scales == '2x':
            return [
                {'scale': 2, 'quality': settings.quality_2x},
            ]
        elif settings.highpixeldensity_scales == '3x':
            return [
                {'scale': 2, 'quality': settings.quality_2x},
                {'scale': 3, 'quality': settings.quality_3x},
            ]
    return []



# -*- coding: utf-8 -*-

from logging import getLogger
from StringIO import StringIO

import struct


log = getLogger(__name__)


def process_tiff(data):
    """handle Tiff Images
    --> Doc http://partners.adobe.com/public/developer/en/tiff/TIFF6.pdf

    """
    content_type = None
    w = -1
    h = -1
    # check for '42' as flag for tiff:
    # TODO: implement correct Image Length and Image Width lookup --> Page 14ff
    # Page 18: Tags:
    # ImageLength: Tag: 257 (101.H) Short or Long
    # ImageWidth: Tag: 256 (100.H) Short or Long

    try:
        # Image File Header (Page 13-14):
        # First 2 Bytes: Determ Byte Order
        # --> II (4949.H) --> little-endian
        # --> MM (4D4D.H) --> big-endian
        # next 2 Bytes always Number: 42
        endian = None
        if data[:2] == 'MM' and struct.unpack('>I', data[2:4])[0] == 42:
            content_type = 'image/tiff'
            endian = '>'  # big-endian encoding for the whole data stream
            log.info('Tiff Image in big-endian encoding')
        elif data[:2] == 'II' and struct.unpack('<I', data[2:4])[0] == 42:
            content_type = 'image/tiff'
            endian = '<'  # little-endian encoding for the whole data stream
            log.info('Tiff Image in little-endian encoding')
        else:
            # not a tiff image
            log.info('Endian or 42 Check failed')

        if endian:
            tiff = StringIO(data)
            tiff.read(4)  # Magic Header, could be skipped, already processed
            offset = struct.unpack_from(endian + 'I', tiff)  # first IFD offset
            b = tiff.read(offset)
            # Process Image File Directory
            while (b and ord(b) != 0xDA):
                field_tag = struct.unpack_from(endian + 'I', tiff)
                field_type = struct.unpack_from(endian + 'I', tiff)
                field_type = translate_field_type.get(field_type, field_type)
                field_value = struct.unpack_from(endian + field_type, tiff)
                if field_tag == '256':  # ImageWidth
                    w = field_value
                elif field_tag == '257':  # ImageLength
                    h = field_value
                    # as fields has to appear in ascending order
                    # we could skip all other fields
                    break
                next_offset = struct.unpack_from(endian + 'I', tiff)
                b.read(next_offset)
    except struct.error:
        pass
    except ValueError:
        pass
    except TypeError:
        pass

    width = int(w)
    height = int(h)
    return content_type, width, height


translate_field_type = {
    """handle Tiff Image File Directory Types
    --> Doc http://partners.adobe.com/public/developer/en/tiff/TIFF6.pdf
    page 14-16
    """
    # TODO: translate to correct python struct mapping
    # TODO: check mappings
    '1': 'I',  # BYTE: 8-bit unsigned Integer
    '2': 'c',  # 'b' 'B'
               # ASCII: 8-bit byte that contains a 7-bit ASCII code
    '3': 'H',  # SHORT: 16-bit (2-byte) unsigned integer
    '4': 'L',  # LONG: 32-bit (4-byte) unsigned integer
    '5': '',  # RATIONAL, two LONGs: the first represents the numerator
              # of a fraction; the second, the donominator
    '6': '',  # SBYTE: An 8-bit signed (twos-complement) integer
    '7': '',  # UNDEFINED
    '8': '',  # SSHORT: A 16-bit (2-byte) signed (twos-complement) integer
    '9': '',   # SLONG: A 32-bit (4-byte) signed (twos-complement) integer
    '10': '',  # SRATIONAL: Two SLONG's (mutator, denominator)
    '11': '',  # FLOAT: Single precision (4-byte) IEEE format.
    '12': '',  # DOUBLE: Double precision (8-byte) IEEE format.
}



# -*- coding: utf-8 -*-
from plone.namedfile.testing import PLONE_NAMEDFILE_FUNCTIONAL_TESTING
from plone.testing import layered

import doctest
import unittest


TEST_FILES = [
    'usage.rst',
    'handler.rst',
    'marshaler.rst',
    'utils.rst',
]


def test_suite():
    return unittest.TestSuite(
        [
            layered(
                doctest.DocFileSuite(
                    testfile,
                    package='plone.namedfile',
                ),
                PLONE_NAMEDFILE_FUNCTIONAL_TESTING
            ) for testfile in TEST_FILES
        ]

    )


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
from plone.testing import Layer
from plone.testing import publisher
from plone.testing import z2
from plone.testing import zca
from plone.testing import zodb
from zope.configuration import xmlconfig


class NamedFileTestLayer(Layer):

    defaultBases = (z2.STARTUP, publisher.PUBLISHER_DIRECTIVES)

    def setUp(self):
        zca.pushGlobalRegistry()

        import plone.namedfile
        xmlconfig.file('testing.zcml', plone.namedfile)

        self['zodbDB'] = zodb.stackDemoStorage(
            self.get('zodbDB'),
            name='NamedFileFixture'
        )

    def tearDown(self):
        # Zap the stacked ZODB
        self['zodbDB'].close()
        del self['zodbDB']

        # Zap the stacked zca context
        zca.popGlobalRegistry()


PLONE_NAMEDFILE_FIXTURE = NamedFileTestLayer()

PLONE_NAMEDFILE_INTEGRATION_TESTING = z2.IntegrationTesting(
    bases=(PLONE_NAMEDFILE_FIXTURE, ),
    name='plone.namedfile:NamedFileTestLayerIntegration',
)

PLONE_NAMEDFILE_FUNCTIONAL_TESTING = z2.FunctionalTesting(
    bases=(PLONE_NAMEDFILE_FIXTURE, ),
    name='plone.namedfile:NamedFileTestLayerFunctional',
)
